#
# Copyright (C) 2012 Marlus Saraiva, Rodrigo Maia
# Copyright (C) 2019 Stefan Eilers
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

require "jasper-rails/version"
require "rails"
require "rjb"
#require "action_controller/metal/responder"

if Mime::Type.lookup_by_extension("pdf").nil?
    Mime::Type.register "application/pdf", :pdf
end

module JasperRails

    class << self
        attr_accessor :config
    end

    module Jasper
        module Rails

            def self.classpath
                classpaths = '.'
                Dir["#{File.dirname(__FILE__)}/java/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end

                Dir["lib/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end
                classpaths
            end

            ##
            ## @brief      Render a PDF file based on jasper file and an optional datasource.
            ##
            ## @param      jasper_file  The Jasper-Report Template (.jrxml)
            ## @param      datasource   The Datasource (JRAbstractTextDataSource)
            ## @param      parameters   The parameters (Ruby-Hash)
            ##
            ## @return     The PDF-Data
            ##
            def self.render_pdf(jasper_file, parameters, datasource = nil, create_pdf_a = false )
                setup_jasper
                # The code below is to workaround declaring constants within methods
                # We would like to delay these till a request is received to workaround Apache Passenger issue
                cLocale                      		= Rjb::import 'java.util.Locale'
                cJasperCompileManager        		= Rjb::import 'net.sf.jasperreports.engine.JasperCompileManager'
                cJasperExportManager         		= Rjb::import 'net.sf.jasperreports.engine.JasperExportManager'
                cJasperFillManager           		= Rjb::import 'net.sf.jasperreports.engine.JasperFillManager'
                cJREmptyDataSource           		= Rjb::import 'net.sf.jasperreports.engine.JREmptyDataSource'
                cHashMap                     		= Rjb::import 'java.util.HashMap'
                cJavaString                  		= Rjb::import 'java.lang.String'

                # PDF/A Export
                cJRPdfExporter 						= Rjb::import 'net.sf.jasperreports.engine.export.JRPdfExporter'
                cSimpleExporterInput         		= Rjb::import 'net.sf.jasperreports.export.SimpleExporterInput'
                cSimpleOutputStreamExporterOutput   = Rjb::import 'net.sf.jasperreports.export.SimpleOutputStreamExporterOutput'
                cSimplePdfExporterConfiguration 	= Rjb::import 'net.sf.jasperreports.export.SimplePdfExporterConfiguration'
                cPdfaConformanceEnum 				= Rjb::import 'net.sf.jasperreports.export.type.PdfaConformanceEnum'
                cByteArrayOutputStream				= Rjb::import 'java.io.ByteArrayOutputStream'
                cDefaultJasperReportsContext		= Rjb::import 'net.sf.jasperreports.engine.DefaultJasperReportsContext'

                options ||= {}
                parameters ||= {}
                jrxml_file  = jasper_file.sub(/\.jasper$/, ".jrxml")

                # Basic setup of JasperReports
                if create_pdf_a
                    jasperReportsContext = cDefaultJasperReportsContext.getInstance();
                    # Overwrite default fonts in order to get an embedded report
                    # TODO: This might change the report. Try following: https://medium.com/@seymorethrottle/jasper-reports-adding-custom-fonts-589b55a52e7c
                    jasperReportsContext.setProperty("net.sf.jasperreports.default.pdf.font.name", "net/sf/jasperreports/fonts/dejavu/DejaVuSans.ttf")
                    jasperReportsContext.setProperty("net.sf.jasperreports.default.pdf.embedded", "true");
                end

                begin
                    # Default report params
                    config = {
                        :report_params=>{
                            "REPORT_LOCALE"    => cLocale.new('de', 'DE'),
                            "XML_LOCALE"       => cLocale.new('de', 'DE'),
                            "XML_DATE_PATTERN" => 'yyyy-MM-dd'
                        }
                    }

                    # Converting default report params to java HashMap
                    jasper_params = cHashMap.new
                    config[:report_params].each do |k,v|
                        jasper_params.put(k, v)
                    end

                    # Convert the ruby parameters' hash to a java HashMap, but keeps it as
                    # default when they already represent a JRB entity.
                    # Pay attention that, for now, all other parameters are converted to string!
                    parameters.each do |key, value|
                        jasper_params.put(cJavaString.new(key.to_s), parameter_value_of(value))
                    end

                    # Compile it, if needed
                    if !File.exist?(jasper_file) || (File.exist?(jrxml_file) && File.mtime(jrxml_file) > File.mtime(jasper_file))
                        cJasperCompileManager.compileReportToFile(jrxml_file, jasper_file)
                    end

                    # Fill the report
                    if datasource
                        jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params, datasource || cJREmptyDataSource.new)
                    else
                        jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params)
                    end

                    if create_pdf_a
                        ## Export as PDF/A with embedded fonts
                        icc_path = "#{File.dirname(__FILE__)}/java/sRGB_v4_ICC_preference_displayclass.icc"
                        exporterConfiguration = cSimplePdfExporterConfiguration.new;
                        exporterConfiguration.isTagged()
                        exporterConfiguration.setPdfaConformance(cPdfaConformanceEnum.PDFA_1A);
                        exporterConfiguration.setIccProfilePath(cJavaString.new(icc_path))

                        #  JRPdfExporter exporter = new JRPdfExporter();
                        byteArrayOutputStream = cByteArrayOutputStream.new
                        exporter = cJRPdfExporter.new()
                        exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                        exporter.setConfiguration(exporterConfiguration)
                        exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                        exporter.exportReport();

                        byteArrayOutputStream.toByteArray()
                    else
                        ## Export as usual PDF-File
                        cJasperExportManager._invoke('exportReportToPdf', 'Lnet.sf.jasperreports.engine.JasperPrint;', jasper_print)
                    end
                rescue Exception=>e
                    if e.respond_to? 'printStackTrace'
                        ::Rails.logger.error e.message
                        e.printStackTrace
                    else
                        ::Rails.logger.error e.message + "\n " + e.backtrace.join("\n ")
                    end
                    raise e
                end
            end

            ##
            ## @brief      Render a Jasper-Report by using a JsonDataSource
            ##
            ## @param      jasper_file  The jasper file (.jrxml)
            ## @param      parameters   The parameters (Ruby-Hash)
            ## @param      json_data    The json data (JSON-String)
            ##
            ## @return     { description_of_the_return_value }
            ##
            def self.render_json_to_pdf(jasper_file, parameters, json_data, create_pdf_a = false)
                setup_jasper

                # Convert json string to byte array.
                jsonDataStream = byteArrayInputStream.new(cJavaString.new(json_data).getBytes());
                data_source = jsonDataSource.new(jsonDataStream)

                render_pdf(jasper_file, parameters, data_source, create_pdf_a)
            end

            ##
            ## @brief      Render a report that contains a datasource.
            ##
            ## @param      jasper_file  The jasper file
            ## @param      parameters   The parameters
            ##
            ## @return     The rendered pdf file.
            ##
            def self.render_datasource_to_pdf(jasper_file, parameters, create_pdf_a = false)
                setup_jasper

                render_pdf(jasper_file, parameters, nil, create_pdf_a)
            end

            private

            def self.setup_jasper
                ENV['CLASS_PATH'] ||= classpath
                ENV['JVM_ARGS'] ||= '-Djava.awt.headless=true,-Xms128M,-Xmx256M'

                Rjb::load( ENV['CLASS_PATH'], ENV['JVM_ARGS'].split(",") ) unless Rjb::loaded?
            end

            # Returns the value without conversion when it's converted to Java Types.
            # When isn't a Rjb class, returns a Java String of it.
            def self.parameter_value_of(param)
                cJavaString                  = Rjb::import 'java.lang.String'
                # Using Rjb::import('java.util.HashMap').new, it returns an instance of
                # Rjb::Rjb_JavaProxy, so the Rjb_JavaProxy parent is the Rjb module itself.
                if param.class.parent == Rjb
                    param
                else
                    cJavaString.new(param.to_s)
                end
            end
        end
    end
end
